/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.hotmail.frojasg1.encrypting.encoderdecoder;

import com.frojasg1.general.GeneralException;
import com.frojasg1.general.HexadecimalFunctions;
import com.frojasg1.general.progress.CancellationException;
import com.frojasg1.general.progress.GeneralUpdatingProgress;
import com.frojasg1.general.progress.UpdatingProgress;
import com.frojasg1.general.progress.OperationCancellation;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.PseudoRandomGenerator;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.PseudoRandomGeneratorException;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.chaoticgenerator.ChaoticGenerator_BigDecimal;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.chaoticgenerator.ChaoticGenerator_BigLong_depretated;
import com.hotmail.frojasg1.encrypting.randomnumbers.RandomSource;
import java.util.Date;

public class SimpleEncoder implements EncoderDecoder
{
	protected byte[] _key = null;
	
	public SimpleEncoder() throws PseudoRandomGeneratorException
	{
		a_generator = new ChaoticGenerator_BigDecimal( 24, 4 );
	}
	
	public SimpleEncoder(int sizeOfNumbers, int numberOfBitsPerIteration ) throws PseudoRandomGeneratorException
	{
		a_generator = new ChaoticGenerator_BigDecimal( sizeOfNumbers, numberOfBitsPerIteration );
	}
	
	public SimpleEncoder( PseudoRandomGenerator generator ) throws PseudoRandomGeneratorException
	{
		a_generator = generator;
	}
	
	@Override
	public int M_getRecommendedNumberOfBytesToInitializeKey()
	{
		return( a_generator.M_getRecommendedNumberOfBytesToInitializeKey() );
	}

	@Override
	public byte[] getKey()
	{
		return( _key );
	}

	@Override
	public void M_initializeKey( byte[] newKey ) throws PseudoRandomGeneratorException
	{
		_key = newKey;
		a_generator.M_initialize(newKey);
	}

	@Override
	public void M_initializeKey( String newHexKey ) throws PseudoRandomGeneratorException
	{
		byte[] byteKey = null;
		try
		{
			byteKey = HexadecimalFunctions.M_convertHexadecimalStringToByteArray(newHexKey);
		}
		catch( GeneralException ex )
		{
			ex.printStackTrace();
			throw( new PseudoRandomGeneratorException( ex.getMessage() ) );
		}
		M_initializeKey( byteKey );
	}

	@Override
	public byte[] M_encode( byte[] input, UpdatingProgress up )
		throws PseudoRandomGeneratorException, CancellationException
	{
		byte[] result = new byte[ input.length ];
		
		if( up != null )
		{
			up.up_childStarts();
		}

		int updatePeriod = 0;
		for( int ii=0; ii<input.length; ii++ )
		{
			result[ii] = (byte)( input[ii] ^ a_generator.M_next() );
			
			if( updatePeriod == 0 )
			{
				if( up != null )
				{
					double progress = ( (double) ii + 1 ) / input.length;
					up.up_updateProgressFromChild( progress );
				}

				updatePeriod = -2000;
			}
			updatePeriod++;
		}

		if( up != null )
		{
			up.up_childEnds();
		}

		return( result );
	}

	@Override
	public byte[] M_decode( byte[] input, UpdatingProgress up )
		throws PseudoRandomGeneratorException, CancellationException
	{
		return( M_encode( input, up ) );
	}

	protected PseudoRandomGenerator a_generator;

	public static void M_testHistogram( int sizeOfNumbers, int bitsPerIteration ) throws PseudoRandomGeneratorException, CancellationException
	{
		RandomSource rs = RandomSource.instance();

		Date time1 = new Date();
		ChaoticGenerator_BigLong_depretated gen = new ChaoticGenerator_BigLong_depretated( sizeOfNumbers, bitsPerIteration );
		SimpleEncoder se = new SimpleEncoder( gen );
		byte[] randomKey = rs.M_getRandomBytes( se.M_getRecommendedNumberOfBytesToInitializeKey() );

		se.M_initializeKey(randomKey );

		byte[] buffer = new byte[256*1024];
		for( int jj=0; jj<buffer.length; jj++ ) buffer[jj]=0;
		byte[] bufferEncoded = se.M_encode(buffer, null);

		int[] histogram = HexadecimalFunctions.M_getHistogram( bufferEncoded );
		System.out.println( String.format( "Test. sizeOfNumbers: %d, bitsPerIteration: %d", sizeOfNumbers, bitsPerIteration ) );
		System.out.println( HexadecimalFunctions.M_IntArrayToString( histogram ) );

		Date time2 = new Date();
		System.out.println( "time in ms " + (time2.getTime() - time1.getTime() ) + "\n\n"  );
	}
	
	public static void M_testHistograms()
	{
		try
		{
			for( int ii=8; ii<=64; ii=ii+8 )
			{
				M_testHistogram( 10, ii );
			}
		}
		catch( Throwable ex )
		{
			ex.printStackTrace();
		}
	}
	
	public static void main( String [] params )
	{
		M_testHistograms();
	}

}
